/*
 *     This program is free software; you can redistribute it
 *     and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation; either
 *     version 2 of the License, or (at your option) any later
 *     version.
 * 
 *     This program is distributed in the hope that it will be
 *     useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more
 *     details.
 * 
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 59 Temple Place - Suite 330, 
 *     Boston, MA  02111, USA.
 * 
 */

/* 
 * Base64 handling elsewhere in mutt should be modified to call
 * these routines. These routines were written because IMAP's
 * AUTHENTICATE protocol required them, and base64 handling
 * elsewhere wasn't sufficiently generic.
 * 
 */

/* 
 * This code is heavily modified from fetchmail (also GPL'd, of
 * course) by Brendan Cully <brendan@kublai.com>.
 * 
 * Original copyright notice:
 * 
 * The code in the fetchmail distribution is Copyright 1997 by Eric
 * S. Raymond.  Portions are also copyrighted by Carl Harris, 1993
 * and 1995.  Copyright retained for the purpose of protecting free
 * redistribution of source. 
 * 
 */

/*
 * This code is from mutt.
 */

int Index_64[128] = [
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,62, -1,-1,-1,63,
    52,53,54,55, 56,57,58,59, 60,61,-1,-1, -1,-1,-1,-1,
    -1, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25,-1, -1,-1,-1,-1,
    -1,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51,-1, -1,-1,-1,-1
];

//#define base64val(c) Index_64[(uint)(c)]

char B64Chars[64] = [
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
  'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
  't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
  '8', '9', '+', '/'
];

enum { BAD = -1 };

/* raw bytes to null-terminated base 64 string */
void mutt_to_base64 (ubyte *outp, ubyte *inp, uint len, uint olen)
{
  while (len >= 3 && olen > 10)
  {
    *outp++ = B64Chars[inp[0] >> 2];
    *outp++ = B64Chars[((inp[0] << 4) & 0x30) | (inp[1] >> 4)];
    *outp++ = B64Chars[((inp[1] << 2) & 0x3c) | (inp[2] >> 6)];
    *outp++ = B64Chars[inp[2] & 0x3f];
    olen  -= 4;
    len   -= 3;
    inp    += 3;
  }

  /* clean up remainder */
  if (len > 0 && olen > 4)
  {
    ubyte fragment;

    *outp++ = B64Chars[inp[0] >> 2];
    fragment = (inp[0] << 4) & 0x30;
    if (len > 1)
      fragment |= inp[1] >> 4;
    *outp++ = B64Chars[fragment];
    *outp++ = (len < 2) ? '=' : B64Chars[(inp[1] << 2) & 0x3c];
    *outp++ = '=';
  }
  *outp = '\0';
}

/* Convert '\0'-terminated base 64 string to raw bytes.
 * Returns length of returned buffer, or -1 on error */
int mutt_from_base64 (char *outp, char *inp)
{
  int len = 0;
  ubyte digit1, digit2, digit3, digit4;

  do
  {
    digit1 = inp[0];
    if (digit1 > 127 || Index_64 [digit1] == BAD)
      return -1;
    digit2 = inp[1];
    if (digit2 > 127 || Index_64 [digit2] == BAD)
      return -1;
    digit3 = inp[2];
    if (digit3 > 127 || ((digit3 != '=') && (Index_64 [digit3] == BAD)))
      return -1;
    digit4 = inp[3];
    if (digit4 > 127 || ((digit4 != '=') && (Index_64 [digit4] == BAD)))
      return -1;
    inp += 4;

    /* digits are already sanity-checked */
    *outp++ = (Index_64[digit1] << 2) | (Index_64[digit2] >> 4);
    len++;
    if (digit3 != '=')
    {
      *outp++ = ((Index_64[digit2] << 4) & 0xf0) | (Index_64[digit3] >> 2);
      len++;
      if (digit4 != '=')
      {
	*outp++ = ((Index_64[digit3] << 6) & 0xc0) | Index_64[digit4];
	len++;
      }
    }
  }
  while (*inp && digit4 != '=');

  return len;
}
