private import obj;
private import objspecs;
private import game;
private import texture;
private import common;
private import mygl;
private import particle;
private import rainbow_color;
private import effect;

abstract class EnemyView : Obj {
    void setXY(float x, float y) {
        x_ = x;
        y_ = y;
    }

    void destroy(float damageAngle) {
    }

    float x_, y_;
}

class ProtonView : EnemyView {
    mixin HasRelativePush;

    this(float x, float y) {
        for (int i = 0; i < NUM; i++) {
            px_[i] = x + rnd(10.0) - 5.0;
            py_[i] = y + rnd(10.0) - 5.0;
            psx_[i] = 0;
            psy_[i] = 0;
            setRndColor(pc_[i]);
        }
    }

    override void move() {
        for (int i = 0; i < NUM; i++) {
            float dx = x_ - px_[i];
            float dy = y_ - py_[i];
            float d = 1.0 / sqrt(dx*dx + dy*dy);
            psx_[i] += dx * d + rnd(1.0) - 0.5;
            psy_[i] += dy * d + rnd(1.0) - 0.5;

            static const int clip_spd = 4;
            if (psx_[i] > clip_spd) psx_[i] = clip_spd;
            else if (psx_[i] < -clip_spd) psx_[i] = -clip_spd;
            if (psy_[i] > clip_spd) psy_[i] = clip_spd;
            else if (psy_[i] < -clip_spd) psy_[i] = -clip_spd;

            px_[i] += psx_[i];
            py_[i] += psy_[i];
        }
    }

    override void draw() {
        GLfloat[] texcoord = Texture.prepare(TextureID.TEX_PARTICLE);

        float tx = 16;
        float ty = 16;

        for (int i = 0; i < NUM; i++) {
            pushMatrix(px_[i], py_[i]);

            glColor4f(pc_[i][0], pc_[i][1], pc_[i][2], 1);

            glBegin(GL_QUADS);
            glTexCoord2f(texcoord[0], texcoord[1]);
            glVertex2f(-tx, -ty);
            glTexCoord2f(texcoord[0], texcoord[3]);
            glVertex2f(-tx, ty);
            glTexCoord2f(texcoord[2], texcoord[3]);
            glVertex2f(tx, ty);
            glTexCoord2f(texcoord[2], texcoord[1]);
            glVertex2f(tx, -ty);
            glEnd();

            glPopMatrix();
        }

        glDisable(GL_TEXTURE_2D);
    }

    override void destroy(float damageAngle) {
        for (int i = 0; i < NUM; i++) {
            Game.obj.addParticle(
                new Particle(
                    px_[i], py_[i],
                    psx_[i]+3*sin(damageAngle), psy_[i]+3*cos(damageAngle),
                    16, pc_[i]));
        }
        Game.obj.addParticle(new ExpEffect(x_, y_));
    }

private:
    static const int NUM = 5;
    float[NUM] px_, py_;
    float[NUM] psx_, psy_;
    float[3][NUM] pc_;
}

class MiddleView : EnemyView {
    mixin HasRelativePush;

    this(float x, float y) {
        angle_ = 0;
        angleSpeed_ = 0;
        angleAccel_ = 0;
    }

    override void move() {
        angleAccel_ += rnd(0.1)-0.05;
        if (angleAccel_ > 1) angleAccel_ = 1;
        else if (angleAccel_ < -1) angleAccel_ = -1;
        angleSpeed_ += angleAccel_;
        if (angleSpeed_ > 10) {
            angleSpeed_ = 10;
            angleAccel_ -= 0.02;
        }
        else if (angleSpeed_ < -10) {
            angleSpeed_ = -10;
            angleAccel_ += 0.02;
        }
        angle_ += angleSpeed_;
    }

    override void draw() {
        pushMatrix();

        glRotatef(angle_, 0, 0, 1);

        GLfloat[] texcoord = Texture.prepare(TextureID.TEX_PARTICLE);

        glRandColor(0.9);

        glBegin(GL_QUADS);
        glTexCoord2f(texcoord[0], texcoord[1]);
        glVertex2f(58, 0);
        glTexCoord2f(texcoord[0], texcoord[3]);
        glVertex2f(14, 15);
        glTexCoord2f(texcoord[2], texcoord[3]);
        glVertex2f(-24, 0);
        glTexCoord2f(texcoord[2], texcoord[1]);
        glVertex2f(14, -15);

        glTexCoord2f(texcoord[0], texcoord[1]);
        glVertex2f(-58, 0);
        glTexCoord2f(texcoord[0], texcoord[3]);
        glVertex2f(-14, 15);
        glTexCoord2f(texcoord[2], texcoord[3]);
        glVertex2f(24, 0);
        glTexCoord2f(texcoord[2], texcoord[1]);
        glVertex2f(-14, -15);

        glTexCoord2f(texcoord[0], texcoord[1]);
        glVertex2f(0, 58);
        glTexCoord2f(texcoord[0], texcoord[3]);
        glVertex2f(15, 14);
        glTexCoord2f(texcoord[2], texcoord[3]);
        glVertex2f(0, -24);
        glTexCoord2f(texcoord[2], texcoord[1]);
        glVertex2f(-15, 14);

        glTexCoord2f(texcoord[0], texcoord[1]);
        glVertex2f(0, -58);
        glTexCoord2f(texcoord[0], texcoord[3]);
        glVertex2f(15, -14);
        glTexCoord2f(texcoord[2], texcoord[3]);
        glVertex2f(0, 24);
        glTexCoord2f(texcoord[2], texcoord[1]);
        glVertex2f(-15, -14);

        glEnd();

        glPopMatrix();
    }

    override void destroy(float damageAngle) {
        for (int i = 0; i < NUM; i++) {
            Game.obj.addParticle(new ExpEffect(x_, y_, 10 - i*13));
/*
            Game.obj.addParticle(
                new Particle(
                    x_, y_,
                    x_+3*sin(damageAngle), y_+3*cos(damageAngle),
                    16));
*/
        }
    }

private:
    static const int NUM = 5;

    float angle_;
    float angleSpeed_;
    float angleAccel_;
}
