private import std.stream;
private import std.string;
private import std.conv;
private import std.math;
private import std.gc;

private import SDL;

private import mygl;
private import sino;
private import player;
private import shot;
private import charactor;
private import myglbf;
private import effect;
private import texture;
private import sound;
private import enemy;
private import common;
private import objmgr;
private import obj;
private import particle;
private import info_draw;
private import viewports;
private import drawobjs;
private import pattern;
private import rainbow_color;
private import reprec;
private import rle;
private import bg_lattice;
private import pversion;

class Game {
private:
    static Game obj_;
public:
    static Game obj() {
        return obj_;
    }

public:
    this(char[] replayFile) {
        replayEndTurn_ = int.max;

        repMode();
        File ifile = new File(replayFile, FileMode.In);
        char[] line;
        enum InputEnc {
            NONE, SIMPLE_RLE
        }
        InputEnc inputEnc = InputEnc.NONE;
        while ((line = ifile.readLine).length != 0) {
            char[][] kv = line.split(`: `);
            char[] k = kv[0];
            char[] v = join(kv[1..length], `: `);
            if (k == `seed`) {
                reprecGenRand.setSeed(toLong(v));
            }
            else if (k == `turn`) {
                replayEndTurn_ = toInt(v);
            }
            else if (k == `mode`) {
                setMode(v);
            }
            else if (k == `inputEnc`) {
                if (v == `simpleRLE`) {
                    inputEnc = InputEnc.SIMPLE_RLE;
                }
            }
            else if (k == `version`) {
                initVersion(v);
            }
        }

        while (!ifile.eof()) {
            char c = ifile.getc();
            reprecBuf ~= c;
        }
        if (inputEnc == InputEnc.SIMPLE_RLE) {
            reprecBuf = rle.decode(reprecBuf);
        }

        init();
    }

    this() {
        setMode(Sino.rc.get(`gameMode`));

        replayEndTurn_ = int.max;
        recMode();

        initVersion();

        init();
    }

    private void setMode(char[] modeStr) {
        if (modeStr == `easy`) {
            mode = Mode.EASY;
        }
        else if (modeStr == `endless`) {
            mode = Mode.ENDLESS;
        }
        else if (modeStr == `normal`) {
            mode = Mode.NORMAL;
        }
    }

    void init() {
        obj_ = this;

        screenWidth = Sino.obj.screenWidth;
        screenHeight = Sino.obj.screenHeight;
        gameWidth = Sino.rc().getInt("gameWidth");
        gameHeight = Sino.rc().getInt("gameHeight");
        gameScreenX = Sino.rc().getInt("gameScreenX");
        gameScreenY = Sino.rc().getInt("gameScreenY");
        gameScreenWidth =
            cast(int)(Sino.rc.getFloat("gameScreenWidth")*screenWidth);
        gameScreenHeight =
            cast(int)(Sino.rc.getFloat("gameScreenHeight")*screenHeight);

        fieldWidth = Sino.rc().getInt("fieldWidth");
        fieldHeight = Sino.rc().getInt("fieldHeight");

        rootObjMgr_ = new ObjMgr!(Obj)("root");

        // game objects

        rootObjMgr_.add(new GameViewport());

        rootObjMgr_.add(new FieldDraw());

//        for (int i = 0; i < 10; i++)
//            rootObjMgr_.add(new BgLines());
        rootObjMgr_.add(new BgLattice());

        player_ = new Player();
        rootObjMgr_.add(player_);

        enemyMgr_ = new ObjMgr!(Enemy)("enemy");
        rootObjMgr_.add(enemyMgr_);

        shotMgr_ = new ObjMgr!(Shot)("shot");
        rootObjMgr_.add(shotMgr_);

        particleMgr_ = new ObjMgr!(Obj)("particle");
        rootObjMgr_.add(particleMgr_);

        // system objects

        rootObjMgr_.add(new SysViewport());

        rootObjMgr_.add(new PsynoDraw());
        rootObjMgr_.add(new MapDraw());
        rootObjMgr_.add(new ScoreDraw());
        rootObjMgr_.add(new PatternDraw());

        turn_ = 0;
        score_ = 0;
        playerLeft_ = 4;

        prepShotNum_ = 0;

        moveWait_ = 0;

        std.gc.fullCollect();
//        std.gc.disable();

        PatternMgr.init();

        Particle.init();

        Sound.obj.init();
        if (mode != Mode.ENDLESS) {
            Sound.obj.loadMusic(`digitalized_heaven.ogg`);
            Sound.obj.playMusic(0);
        }
        else {
            Sound.obj.loadMusic(`digitalized_heaven_quietpraymix.ogg`);
            Sound.obj.playMusic(-1);
        }
    }

    void saveReplay(Stream ofile) {
        ofile.writeLine("patterns: " ~ tostr(destructPatternNum_));
        ofile.writeLine("turn: " ~ tostr(turn_));
        ofile.writeLine("seed: " ~ tostr(reprecGenRand.seed));
        static const char[][] mode2str = [ `easy`, `normal`, `endless` ];
        ofile.writeLine("mode: " ~ mode2str[mode]);
        ofile.writeLine("inputEnc: simpleRLE");
        ofile.writeLine("");
        foreach (char c; rle.encode(reprecBuf)) {
            ofile.write(c);
        }
    }

    void run() {
        if (isEnd_) {
            endCnt_++;
        }
        else {
            reprecNext();
        }

        if (replayEndTurn_ <= turn_) {
            isEnd_ = true;
        }

        if (!(turn_ % 62)) {
            scoreStat_ ~= score_;
/*
            char[] msg = "GTURN: " ~ tostr(turn_);
            msg ~= "  ENEMY: " ~ tostr(enemyMgr_.objNum());
            msg ~= "  SHOT: " ~ tostr(shotMgr_.objNum());
            msg ~= "  PARTICLE: " ~ tostr(particleMgr_.objNum());
            Log.info(msg);
*/
        }

        addPatternWait_--;
        if (addPatternWait_ < 0 && patterns_.length < 5 &&
            player_.comboCnt() == 0) {
            if (!PatternMgr.addNextPattern() && enemyMgr_.objNum == 0) {
                isEnd_ = true;
            }
        }
/*
        if (enemyMgr_.objNum() < 5 || rnd(300) == 0) {
            PatternMgr.addRandomPattern();
        }
*/
/*
        if (rnd(50) == 0) {
            if (enemyMgr_.objNum() < 30) {
                float x = rnd(fieldWidth);
                float y = rnd(fieldHeight);
                if (!isInside(x, y)) {
                    enemyMgr_.add(new Enemy(x, y));
                }
            }
        }
*/

        moveWait_--;
        if (moveWait_ < 0 || (moveWait_ & 1)) {
            rootObjMgr_.move();
        }
        else {
            particleMgr_.move();
        }

        for (int i = 0; i < prepShotNum_; i++) {
            with (prepShots_[i]) {
                shotMgr_.add(new Shot(x, y, sx, sy, col));
            }
        }
        prepShotNum_ = 0;

        prepEnemyCnt_--;
        if (prepEnemyCnt_ == 0) {
            Sound.obj.playChunk(Se.ENEMY_APPEAR);

            foreach (Enemy e; prepEnemies_) {
                addEnemy(e);
            }
            prepEnemies_.length = 0;
        }
        else if (prepEnemyCnt_ > 0 && (prepEnemyCnt_ & 1)) {
            foreach (Enemy e; prepEnemies_) {
                float[3] col;
                setRndColor(col);
                Particle p = new Particle(e.x(), e.y(), 0, 0, 7, col);
                addParticle(p);
            }
        }

        if (!Sino.obj().timer().shouldSkip()) {
            draw();

            if (isEnd_) {
                glDisable(GL_TEXTURE_2D);
                glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                glColor4f(0, 0, 0, endCnt_/60.0);
                glBegin(GL_QUADS);
                glVertex2f(480, 0);
                glVertex2f(480, 480);
                glVertex2f(0, 480);
                glVertex2f(0, 0);
                glEnd();
                glBlendFunc(GL_SRC_ALPHA, GL_ONE);
            }
        }

        if (mode != Mode.ENDLESS) {
            if (!isEnd_ && turn_ < 16000) {
                turn_++;

                int lm = 0;
                if (turn_ == 5200) lm = 3;
                else if (turn_ == 8800) lm = 2;
                else if (turn_ == 12400) lm = 1;
                if (lm) {
                    addParticle((new StringEffect(
                                     80, 300,
                                     "Last " ~ tostr(lm) ~ " min.",
                                     14, 18, 1,1,1, 90)
                                    ).center.speed(2.66667, 0).absolute);
                }
                int lf = 15999-turn_;
                if (lf < 610 && lf % 60 == 0) {
                    addParticle((new StringEffect(
                                     80, 300,
                                     tostr(lf/60) ~ " sec.",
                                     14, 18, 1,1,1, 90)
                                    ).center.speed(2.66667, 0).absolute);
                }
            }
            else {
                isEnd_ = true;
            }
        }
        else { // ENDLESS
            turn_++;
        }
    }

    void draw() {
        glClear(GL_COLOR_BUFFER_BIT);

        rootObjMgr_.draw();

    }

    void convPos(inout float x, inout float y) {
        x -= scrOffsetX;
        y -= scrOffsetY;
//        x = player_.x() - (gameWidth>>1) - x;
//        y = player_.y() - (gameHeight>>1) - y;
    }

    bool isInside(float x, float y, int off) {
        convPos(x, y);
        return x > -off
            && y > -off
            && x < gameWidth+off
            && y < gameHeight+off;
    }
    bool isInside(float x, float y) {
        return isInside(x, y, 0);
    }

    void updatePlayerPos(float x, float y) {
        scrOffsetX = cast(int)x - (gameWidth >> 1);
        scrOffsetY = cast(int)y - (gameHeight >> 1);
    }

    private void drawResultStringC(char[] str, int y) {
        drawString(str, (640-12*str.length)/2, y, 12, 16, 1, 1, 1);
    }
    private void drawResultStringL(char[] str, int y) {
        drawString(str, 200, y, 12, 16, 1, 1, 1);
    }
    private void drawResultStringR(char[] str, int y) {
        drawString(str, 440-12*str.length,
                   y, 12, 16, 1, 1, 1);
    }

    bit isEnd() {
        return isEnd_ && endCnt_ > 60;
    }

    void end() {
        isEnd_ = true;
    }

    Player player() {
        return player_;
    }

    Enemy[] enemies() {
        return enemyMgr_.objs;
    }

    Shot[] shots() {
        return shotMgr_.objs;
    }

    int turn() { return turn_; }

    void addParticle(Obj p) {
        if (particleMgr_.objNum < 200)
            particleMgr_.add(p);
    }

    void addShot(float x, float y, float sx, float sy, float[3] col) {
        if (prepShotNum_ == PREP_SHOT_NUM ||
            shotMgr_.objNum() + prepShotNum_ > 400) {
            //Log.warn("maximam appear shot");
            return;
        }

//        shotMgr_.add(new Shot(x, y, sx, sy, col));

        prepShots_[prepShotNum_].x = x;
        prepShots_[prepShotNum_].y = y;
        prepShots_[prepShotNum_].sx = sx;
        prepShots_[prepShotNum_].sy = sy;
        prepShots_[prepShotNum_].col[0] = col[0];
        prepShots_[prepShotNum_].col[1] = col[1];
        prepShots_[prepShotNum_].col[2] = col[2];
        prepShotNum_++;
    }

    void addEnemy(Enemy e) {
        enemyMgr_.add(e);
    }

    void addEnemyPattern(int patternNum, Enemy[] enemies) {
        addPatternWait_ = 180;

        if (patternNum in patterns_) return;

        patterns_[patternNum] = enemies.length;

        prepEnemyCnt_ = 60;
        prepEnemies_ = enemies;
    }

    void delEnemy(int pattern) {
        if (--patterns_[pattern] == 0) {
            int i = 0;
            foreach (int k, int v; patterns_) {
                if (k == pattern) break;
                i++;
            }
            addParticle((new StringEffect(
                             400, 340+i*22,
                             "Pattern " ~ tostr(pattern+1) ~ " closed.",
                             12, 16, 1,1,1, 90)
                            ).center.speed(-2.66667, 0).absolute);

            delete patterns_[pattern];
            destructPatternNum_++;
            if (destructPatternNum_ % 3 == 0) {
                playerLeft_++;

                Sound.obj.playChunk(Se.PLAYER_EXTEND);

                addParticle((new StringEffect(
                                 80, 50,
                                 "Player extended.",
                                 14, 18, 1,1,1, 90)
                                ).center.speed(2.66667, 0).absolute);

                if (playerLeft_ == 10) {
                    addParticle((new StringEffect(
                                     80, 255,
                                     "perfect play.",
                                     14, 18, 1,1,1, 90)
                                    ).center.speed(2.66667, 0).absolute);
                }
            }
        }
    }

    void damagePlayer() {
        if (playerLeft_ == 0) {
            isEnd_ = true;
        }
        else {
            playerLeft_--;

            if (playerLeft_ == 0) {
                addParticle((new StringEffect(
                                 80, 255,
                                 "last player.",
                                 14, 18, 1,1,1, 90)
                                ).center.speed(2.66667, 0).absolute);
            }
        }

        if (versionNum > 10) {
            foreach (Shot s; shots()) {
                with (s) {
                    addParticle(new Particle(x(), y(), sx(), sy(), 10, col));
                    vanish();
                }
            }
        }

        Sound.obj.playChunk(Se.PLAYER_DEAD);
    }

    int[int] patterns() { return patterns_; }

    void setMoveWait(int w) {
        moveWait_ = w;
    }

    int objectNum() {
        return enemyMgr_.objNum + shotMgr_.objNum + particleMgr_.objNum;
    }

    int score() { return score_; }

    void addScore(int s) { score_ += s; }

    bool isCleared() {
        return mode != Mode.ENDLESS && destructPatternNum_ == 20;
    }

public:
    // for scene_result

    int playerLeft() { return playerLeft_; }

    int[] scoreStat() { return scoreStat_; }

private:
    bit isEnd_;

    int turn_;

    int score_;

public:
    int screenWidth;
    int screenHeight;

    int gameWidth;
    int gameHeight;
    int gameScreenX;
    int gameScreenY;
    int gameScreenWidth;
    int gameScreenHeight;

    int fieldWidth;
    int fieldHeight;

    enum Mode { EASY, NORMAL, ENDLESS }
    Mode mode;

private:
    int scrOffsetX;
    int scrOffsetY;

    struct PrepShot {
        float x, y, sx, sy;
        float[3] col;
    }
    static const int PREP_SHOT_NUM = 50;
    PrepShot[PREP_SHOT_NUM] prepShots_;
    int prepShotNum_;

    Enemy[] prepEnemies_;
    int prepEnemyCnt_;

    int moveWait_;

    int replayEndTurn_;

    int playerLeft_;
    int destructPatternNum_;

    int[] scoreStat_;

    int endCnt_;

private:
    ObjMgr!(Obj) rootObjMgr_;

    Player player_;
    ObjMgr!(Enemy) enemyMgr_;
    ObjMgr!(Shot) shotMgr_;
    ObjMgr!(Obj) particleMgr_;

    int addPatternWait_;
    int[int] patterns_;

};
