private import std.string;
private import std.math;

private import common;
private import sino;
private import log;
private import charactor;
private import shot;
private import sound;
private import texture;
private import game;
private import particle;
private import objspecs;
private import enemy;
private import mygl;
private import thunder;
private import effect;
private import myglbf;
private import sound;
private import rainbow_color;
private import reprec;
private import pversion;

class Player : public Charactor {
    mixin HasRelativePush;
    mixin HasClipping;

public:
    this() {
        state_ = State.FREE;

        x_ = Sino.rc().getFloat("playerX");
        y_ = Sino.rc().getFloat("playerY");
        sx_ = 0;
        sy_ = 0;

        playerSpeed_ = Sino.rc().getFloat("playerSpeed");
        playerSpeed2_ = Sino.rc().getFloat("playerSpeed2");

        size_ = Sino.rc().getInt("playerSize");
        playerFieldSize_ = Sino.rc().getInt("playerFieldSize");
        playerFieldSize2_ = (playerFieldSize_ * playerFieldSize_) << 2;

        screenWidth_ = Sino.rc().getInt("fieldWidth") - size_;
        screenHeight_ = Sino.rc().getInt("fieldHeight") - size_;

        lastAngle_ = 0;
        holdCnt_ = 0;
    }

    override void move() {
        // @@@ for debug
/*
        if (Pad_GetButtonPressed(PadKey.PADK_q)) {
            Game.obj.end();
        }
*/

/* for state debugging
        static int prSt = -1;
        if (prSt != state_) {
            Log.info(`state changed: ` ~ stateStr[state_]);
            prSt = state_;
        }
*/

        int axis = reprecAxis();
        int aButton = reprecAButton();
        int bButton = reprecBButton();
        int aButtonPushed = reprecAButtonPushed();

        holdCnt_--;
        if (holdCnt_ == 0) {
            state_ = State.FREE;
        }
        swordCnt_--;

        // update nearest
        float disMin = float.max;
        if (isQdInvisible()) {
            static float prevAngle = float.nan;

            if (nearest_ === null ||
                (prevAngle != a_ && state_ == State.ATACKED)) {
                prevAngle = a_;

                if (isnan(a_)) {
                    foreach (Enemy e; Game.obj.enemies()) {
                        float d2 = distance2(x_, y_, e.x, e.y);
                        float reach = playerSpeed_*3*60;
                        if (d2 > reach*reach) continue;
                        if (d2 < disMin) {
                            disMin = d2;
                            nearest_ = e;
                        }
                    }
                }
                else {
                    foreach (Enemy e; Game.obj.enemies()) {
                        float d2 = distance2(x_, y_, e.x, e.y);
                        float reach = playerSpeed_*3*60;
                        if (d2 > reach*reach) continue;

                        float a = fabs(relAngle(a_ - angle(x_, y_, e.x, e.y)));

                        if (a < disMin) {
                            disMin = a;
                            nearest_ = e;
                        }
                    }
                }
            }
        }
        else {
            nearest_ = null;
            foreach (Enemy e; Game.obj.enemies()) {
                if (!Game.obj.isInside(e.x, e.y)) continue;
                float d2 = distance2(x_, y_, e.x, e.y);
                if (d2 < disMin) {
                    disMin = d2;
                    nearest_ = e;
                }
            }
        }

        void setSpeed(float base, int axis) {
            float cross = base / SQRT_2;

            sx_ = sy_ = 0;
            if (axis == PAD_AXIS_UP) sy_ = -base;
            else if (axis == PAD_AXIS_RIGHT) sx_ = base;
            else if (axis == PAD_AXIS_DOWN) sy_ = base;
            else if (axis == PAD_AXIS_LEFT) sx_ = -base;
            else if (axis == PAD_AXIS_UPRIGHT) {
                sx_ = cross;
                sy_ = -cross;
            }
            else if (axis == PAD_AXIS_DOWNRIGHT) {
                sx_ = cross;
                sy_ = cross;
            }
            else if (axis == PAD_AXIS_DOWNLEFT) {
                sx_ = -cross;
                sy_ = cross;
            }
            else if (axis == PAD_AXIS_UPLEFT) {
                sx_ = -cross;
                sy_ = -cross;
            }
        }

        void atackCommon() {
            if (nearest_ === null) return;

            float d2 = distance2(x_, y_, nearest_.x, nearest_.y);
            float a = angle(x_, y_, nearest_.x, nearest_.y);
            float swWidth = PI * 0.12;
            if (state_ == State.SQDATACK) swWidth = PI * 0.2;
            else if (state_ == State.QDATACK) swWidth = PI * 0.08;
            if ((d2 < 100 ||
                 (d2 < SWORD_LEN*SWORD_LEN &&
                  fabs(relAngle(a-swordAngle_)) < swWidth)) ||
                (state_ == State.ROLLATACK && d2 < 10000))
            {
                int score = 0;

                nearest_.damage(comboCnt_+1, a);

                float rs() { return rnd(20.0)-10.0; }

                if (!nearest_.alive) {
                    foreach (Shot s; Game.obj.shots()) {
                        if (!s.alive()) continue;

                        // @@@ ԥڤǥɤ񤯤
                        if (distance2(nearest_.x, nearest_.y,
                                      s.x, s.y) < 10000) {
                            s.die();

                            float[3] col;
                            setRndColor(col);
                            Game.obj.addParticle(
                                new Thunder(s.x(), s.y(), rs(), rs(),
                                            col, null));

                            vanishShotCnt_++;

                            if (vanishShotCnt_ % 10 == 0) {
                                Game.obj.addParticle(new StringEffect(
                                                         s.x(), s.y(),
                                                         tostr(vanishShotCnt_),
                                                         14, 18, 1, 1, 1));
                            }

                            score += (vanishShotCnt_ > 100)
                                ? 1000 : vanishShotCnt_*10;
                        }
                    }

                    if (state_ == State.ATACKED) {
                        holdCnt_ += 10;
                    }
                    else {
                        state_ = State.ATACKED;
                    }

                    comboCnt_++;

                    for (int i = 0; i < comboCnt_*3; i++) {
                        float[3] col;
                        setRndColor(col);
                        Game.obj.addParticle(
                            new Thunder(x_, y_, rs(), rs(), col, null));
                    }

/* for score check
                    static int vscore = 0;
                    vscore += score * comboCnt_;
                    Log.info(tostr(vscore));
*/

                    score += nearest_.score;

/*
                    int sizeAdd = (comboCnt_ > 20) ? 20 : comboCnt_;
                    Game.obj.addParticle(new StringEffect(
                                             nearest_.x(), nearest_.y(),
                                             "X"~tostr(comboCnt_),
                                             20+sizeAdd, 30+sizeAdd*2,
                                             1, 1, 1));
*/

                    int fscore = score * comboCnt_;
                    comboScore_ += fscore;
                    Game.obj.addScore(fscore);

                    Log.info("score: " ~ tostr(score) ~ "*" ~ tostr(comboCnt_)
                             ~ " = " ~ tostr(fscore)
                             ~ " (vanished: " ~ tostr(vanishShotCnt_) ~ ")" );
                }

                Game.obj.setMoveWait(14);

                Sound.obj.playChunk(Se.SMALL_ENEMY_DESTROY);

/*
  for (int i = 0; i < cnt; i++) {
  float[3] col;
  setRndColor(col);
  Game.obj.addParticle(
  new Thunder(x_, y_, rs(), rs(), col, null));
  }
*/
            }
        }

        switch (cast(int)state_) {
        case State.ATACKED:
            if (bButton) {
                holdCnt_ = 60;
                state_ = State.SQD;
            }
            else if (aButtonPushed /*&& holdCnt_ < 25*/) {
                holdCnt_ = 30;
                swordCnt_ = SWORD_CNT;
                state_ = State.ROLLATACK;
                swordAngle_ = lastAngle_ + PI;
                swordSpeed_ = - 2 * PI / SWORD_CNT;

                if (nearest_) {
                    float a = angle(x_, y_, nearest_.x, nearest_.y);
                    sx_ = playerSpeed*7*sin(a);
                    sy_ = playerSpeed*7*cos(a);
                }
            }

        case State.QDATACK:
        case State.SQDATACK:
        case State.ROLLATACK:
            sx_ *= 0.9;
            sy_ *= 0.9;

        case State.ATACK:
            swordAngle_ += swordSpeed_;

            if (swordCnt_ > 0) atackCommon();

            break;

        case State.SQD:
        case State.QD:
            if (nearest_ === null) {
                state_ = State.FREE;
                break;
            }

            if (distance2(x_, y_, nearest_.x, nearest_.y) < 100) {
                state_ = State.FREE;

                break;
            }

            float a = angle(x_, y_, nearest_.x, nearest_.y);
            if (state_ == State.QD) {
                sx_ = playerSpeed*3*sin(a);
                sy_ = playerSpeed*3*cos(a);
            }
            else {
                sx_ = playerSpeed*3*sin(a);
                sy_ = playerSpeed*3*cos(a);
            }

            if (aButton) {
                swordAngle_ = a - 0.5 * QDSWORD_ANGLE;
                swordSpeed_ = QDSWORD_ANGLE / SWORD_CNT;
                holdCnt_ = 30;
                swordCnt_ = SWORD_CNT;
                if (state_ == State.QD) {
                    state_ = State.QDATACK;
                }
                else {
                    state_ = State.SQDATACK;
                }
            }

            break;

        case State.ND:
            setSpeed(playerSpeed*2, ndAxis_);

            break;

        case State.FREE:
            if (aButton) {
                holdCnt_ = 30;
                swordCnt_ = SWORD_CNT;
                state_ = State.ATACK;
                swordAngle_ = lastAngle_ - 0.5 * SWORD_ANGLE;
                swordSpeed_ = SWORD_ANGLE / SWORD_CNT;
            }
            else if (bButton) {
                if (axis == PAD_AXIS_NONE) {
                    holdCnt_ = 20;
                    state_ = State.QD;
                }
                else {
                    holdCnt_ = 20;
                    state_ = State.ND;
                    ndAxis_ = axis;
                }
            }

        case State.DAMAGE:
            comboEnd();

            float base;
            if (aButton) base = playerSpeed2_;
            else base = playerSpeed_;
            setSpeed(base, axis);

            if (versionNum >= 30) {
                if (state_ == State.DAMAGE &&
                    bButton && axis != PAD_AXIS_NONE)
                {
                    damageDashCnt_ = 20;
                    state_ = State.DAMAGEND;
                    ndAxis_ = axis;
                }
            }

            break;

        case State.DAMAGEND:
            assert(versionNum >= 30);

            setSpeed(playerSpeed*2, ndAxis_);

            damageDashCnt_--;
            if (damageDashCnt_ == 0) {
                state_ = State.DAMAGE;
            }

            break;
        }

        super.move();

        clipPos();

        Game.obj.updatePlayerPos(x_, y_);

        if (isQdInvisible()) {
            float[3] red;
            red[0] = 1;
            red[1] = 0.6;
            red[2] = 0.6;
            for (int i = 0; i < 2; i++)
                Game.obj.addParticle(
                    new Particle(x_, y_,
                                 rnd(4.0)-2.0, rnd(4.0)-2.0,
                                 9, red));
        }
        else if (isDash() || isDamage()) {
            Game.obj.addParticle(
                new Particle(x_, y_,
                             -sx_*(rnd(1.0)+0.5), -sy_*(rnd(1.0)+0.5)));
        }
        else if (turn_ & 1) {
            Game.obj.addParticle(
                new Particle(x_, y_,
                             -sx_*(rnd(1.0)+1.0), -sy_*(rnd(1.0)+1.0)));
        }

        static const float[] axis2angle = [
            float.nan, std.math.PI, 0, float.nan,
            -std.math.PI*0.5, -std.math.PI*0.75, -std.math.PI*0.25, float.nan,
            std.math.PI*0.5, std.math.PI*0.75, std.math.PI*0.25,
            float.nan, float.nan, float.nan, float.nan, float.nan
            ];
        a_ = axis2angle[axis];
        //s_ = base;

        if (!isnan(a_)) lastAngle_ = a_;

        // ɤȽ⤳Ǥäޤ
        if (isDamage()) {
        }
        else if (isQdInvisible()) {
            foreach (Shot s; Game.obj.shots()) {
                if (distance2(x_, y_, s.x, s.y) < size_*size_) {
                    s.die();
                    /* 
                    float rs() { return rnd(20.0)-10.0; }
                    float[3] col;
                    setRndColor(col);
                    Game.obj.addParticle(
                        new Thunder(x_, y_, rs(), rs(), col, null));
                    */
                }
            }
        }
        else {
            foreach (Shot s; Game.obj.shots()) {
                if (distance2(x_, y_, s.x, s.y) < 10) {
                    state_ = State.DAMAGE;
                    sx_ = 0;
                    sy_ = 0;
                    holdCnt_ = 180;

                    Game.obj.damagePlayer();

                    for (int i = 0; i < 30; i++) {
                        float[3] col;
                        setRndColor(col);
//                        col[0] = 1; col[1] = 1; col[2] = 1;
                        Game.obj.addParticle(
                            new RevParticle(x_, y_,
                                            rnd(20.0)-10.0, rnd(20.0)-10.0,
                                            13, col));
                    }
/*
                    for (int i = 0; i < 5; i++) {
                        Game.obj.addParticle(new ExpEffect(x_, y_, 10 - i*4));
                    }
*/

                    break;
                }
            }
        }

        if (nearest_ && !nearest_.alive) {
            nearest_ = null;
        }
    }

    float playerSpeed() { return playerSpeed_; }
    float playerSpeed2() { return playerSpeed2_; }

    int screenWidth() { return screenWidth_; }
    int screenHeight() { return screenHeight_; }

    void damage() {
    }

    int comboCnt() { return comboCnt_; }

    bool isQdInvisible() {
        return state_ == State.ATACKED || state_ == State.ROLLATACK ||
            state_ == State.SQD || state_ == State.SQDATACK;
    }
    bool isDash() {
        return state_ == State.SQD || state_ == State.QD || state_ == State.ND;
    }
    bool isAtack() {
        return state_ == State.ATACK || state_ == State.ATACKED ||
            state_ == State.QDATACK || state_ == State.SQDATACK ||
            state_ == State.ROLLATACK;
    }
    bool isDamage() {
        return state_ == State.DAMAGE || state_ == State.DAMAGEND;
    }

    void comboEnd() {
        if (comboCnt_ >= 10) {
            Game.obj.addParticle((new StringEffect(
                                      400, 150,
                                      tostr(comboCnt_) ~ " hit combo. " ~
                                      tostr(comboScore_) ~ " point.",
                                      16, 22, 1,1,1, 290)
                                     ).center.speed(-1, 0).absolute);
        }

        comboCnt_ = 0;
        comboScore_ = 0;
        vanishShotCnt_ = 0;
    }

    override void draw() {
        pushMatrix();

        float tx = size_;
        float ty = size_;
        GLfloat[] texcoord;

        if (comboCnt_ > 0) {
            int sizeAdd = (comboCnt_ > 20) ? 20 : comboCnt_;
            drawString("X"~tostr(comboCnt_), 0, 0,
                       20+sizeAdd, 30+sizeAdd, 1,1,1);
        }

        if (!isDamage()) {
            texcoord = Texture.prepare(TextureID.TEX_PARTICLE);

            glColor4f(1, 1, 1, 1);

            glBegin(GL_QUADS);
            glTexCoord2f(texcoord[0], texcoord[1]);
            glVertex2f(-tx, -ty);
            glTexCoord2f(texcoord[0], texcoord[3]);
            glVertex2f(-tx, ty);
            glTexCoord2f(texcoord[2], texcoord[3]);
            glVertex2f(tx, ty);
            glTexCoord2f(texcoord[2], texcoord[1]);
            glVertex2f(tx, -ty);
            glEnd();
        }

        if (state_ == State.FREE || isQdInvisible()) {
            texcoord = Texture.prepare(TextureID.TEX_BARRIER);

            if (state_ == State.FREE) {
                glColor4f(0.9, 0.9, 1.0, 0.5);
            }
            else {
                glColor4f(1.0, 0.9, 0.9, 1.0);
            }

            tx = 23;
            ty = 23;
            float rs() { return rnd(2.0) - 1.0; }
            glBegin(GL_QUADS);
            glTexCoord2f(texcoord[0], texcoord[1]);
            glVertex2f(-tx-rs(), -ty-rs());
            glTexCoord2f(texcoord[0], texcoord[3]);
            glVertex2f(-tx-rs(), ty+rs());
            glTexCoord2f(texcoord[2], texcoord[3]);
            glVertex2f(tx+rs(), ty+rs());
            glTexCoord2f(texcoord[2], texcoord[1]);
            glVertex2f(tx+rs(), -ty-rs());
            glEnd();
        }

        glPopMatrix();

        glDisable(GL_TEXTURE_2D);

        glBegin(GL_LINES);

        glColor4f(1, 0, 0, 0.5);

        foreach (Enemy e; Game.obj.enemies()) {
            if (e === nearest_) {
                static const int PERIOD = 4;
                int r = turn_ % (PERIOD*3);
                if (r < PERIOD) {
                    glColor4f(1, 0, 0, 0.9);
                }
                else if (r < PERIOD*2) {
                    glColor4f(0, 0, 1, 0.9);
                }
                else {
                    glColor4f(1, 1, 1, 0.9);
                }
            }
            else {
                float d = distance(x_, y_, e.x, e.y);
                if (d > 1000) d = 1000;
                glColor4f(0.001*(1000-d), 0, 0.001*d, 0.7-0.0003*d);
            }
            myglClippedLine2(x_, y_, e.x, e.y);
        }

        glEnd();

        if (isAtack() && swordCnt_ > 0) {
            GLfloat[] texcoord = Texture.prepare(TextureID.TEX_SWORD);

            float x = x_;
            float y = y_;
            Game.obj.convPos(x, y);

            float sn = sin(swordAngle_);
            float cs = cos(swordAngle_);

            float sx = x + SWORD_LEN*sn;
            float sy = y + SWORD_LEN*cs;

            float c = 1.0 - comboCnt_ * 0.05;
            if (c < 0) c = 0;
            glColor3f(1, c, c);

            static const float WIDTH = 5.5;

            glBegin(GL_QUADS);
            glTexCoord2f(texcoord[0], texcoord[1]);
            glVertex2f(x-cs*WIDTH, y-sn*WIDTH);
            glTexCoord2f(texcoord[0], texcoord[3]);
            glVertex2f(x+cs*WIDTH, y+sn*WIDTH);
            glTexCoord2f(texcoord[2], texcoord[3]);
            glVertex2f(sx+cs*WIDTH, sy+sn*WIDTH);
            glTexCoord2f(texcoord[2], texcoord[1]);
            glVertex2f(sx-cs*WIDTH, sy-sn*WIDTH);
            glEnd();

/*
            glLineWidth(11);

            glBegin(GL_LINES);
            glTexCoord2f(texcoord[0], texcoord[1]);
            glVertex2f(x, y);
            glTexCoord2f(texcoord[2], texcoord[3]);
            glVertex2f(sx, sy);
            glEnd();
*/
        }
        glDisable(GL_TEXTURE_2D);
    }

private:
    enum State {
        FREE, ATACK, SQD, QD, ND, DAMAGE, QDATACK,
        ATACKED, SQDATACK, ROLLATACK, DAMAGEND
    }
    State state_;
    static const char[][] stateStr = [
        `FREE`, `ATACK`, `SQD`, `QD`, `ND`, `DAMAGE`, `QDATACK`,
        `ATACKED`, `SQDATACK`, `ROLLATACK`
        ];
private:
    float playerSpeed_;
    float playerSpeed2_;

    int color_;
    int opposite_;

    int playerFieldSize_;
    int playerFieldSize2_;

    int screenWidth_;
    int screenHeight_;

    char[] name_;

    float lastAngle_;

    // Ӥߥ
    int holdCnt_;
    int swordCnt_;
    int damageDashCnt_;

    Enemy nearest_;

    // for state
    int ndAxis_;
    float swordAngle_;
    float swordSpeed_;
    int comboCnt_;
    int comboScore_;
    int vanishShotCnt_;

    static const int SWORD_LEN = 100;
    static const int SWORD_CNT = 15;
    static const float SWORD_ANGLE = PI * 0.8;
    static const float QDSWORD_ANGLE = PI * 0.3;
}
