private import game;
private import scene;
private import mygl;
private import myglbf;
private import sound;
private import common;
private import pversion;
private import fnenc;
private import sino;
private import rc;
private import result_display;
private import http;
private import http_thread;

private import SDL_pad;

private import std.stream;
private import std.date;
private import std.string;

class SceneResult : Scene {
public:
    this(Game game) {
        Log.info("scene result");
        Sound.obj.fadeMusic();

        game_ = game;

        state_ = State.RESULT;

        result_ = new ResultDisplay(`Result`);
        with (result_) {
            gameScore = game.score();
            int secScore = (game.mode == Game.Mode.EASY) ? 500 : 1000;
            turnScore = game.isCleared() * (16000-game.turn()) * secScore;
            clearScore = game.isCleared() * 1000000;
            leftScore = game.playerLeft() * 200000;
            nomissScore = (game.playerLeft() == 10) * 1000000;
            allScore = gameScore+turnScore+clearScore+leftScore+nomissScore;

            endTurn = game.turn;

            if (game.mode == Game.Mode.ENDLESS) {
                if (game.scoreStat.length > 240) {
                    scoreStat.length = 240;
                    for (int i = 0; i < 240; i++) {
                        scoreStat[i] =
                            game.scoreStat[i*game.scoreStat.length/240];
                    }
                }
                else {
                    scoreStat = game.scoreStat;
                }
            }
            else {
                scoreStat = game.scoreStat;
                for (int i = 0; i < (15000-endTurn)/62-1; i++) {
                    scoreStat ~=
                        scoreStat[length-1]+game.isCleared()*62*secScore;
                }
                scoreStat ~= gameScore+turnScore;
                scoreStat ~= allScore;
            }
        }
        result_.init();

        name_ = Sino.rc.hasKey("name") ? Sino.rc.get("name") : "";
    }

    bit isEnd() {
        return isEnd_;
    }

    Scene getNext() {
        return new SceneSelect();
    }

    void run() {
        glClear(GL_COLOR_BUFFER_BIT);

        draws_.draw();

        if (state_ == State.NAME || state_ == State.SEND ||
            state_ == State.SENDING || state_ == State.SENDED)
        {
            doNameEntry();
            return;
        }

        result_.move();
        result_.draw();

        if (result_.turn > 250) {
            if (state_ == State.RESULT) {
                if (Pad_GetButtonPushed(PAD_A_BUTTON) ||
                    Pad_GetButtonPushed(PAD_B_BUTTON))
                {
                    state_ = State.SAVE;
                    yes_ = true;
                    stateStartTurn_ = Sino.obj.timer.getTurn;
                }
            }
            else if (state_ == State.SAVE) {
                doSaveReplay();
            }
        }
        else {
            if (Pad_GetButtonPressed(PAD_A_BUTTON)) {
                for (int i = 0; i < 4; i++) {
                    result_.move();
                }
            }
        }
    }

    private void doMsg(char[] msg) {
        glDisable(GL_TEXTURE_2D);

        float elapsed = Sino.obj.timer.getTurn - stateStartTurn_;
        if (elapsed > 30) elapsed = 30;

        float x1 = 50;
        float x2 = 430;
        float y1 = 230;
        float y2 = 270;

        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glColor4f(0, 0, 0, elapsed/35);
        glBegin(GL_QUADS);
        glVertex2f(x1, y1);
        glVertex2f(x2, y1);
        glVertex2f(x2, y2);
        glVertex2f(x1, y2);
        glEnd();
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        float off = 30-elapsed;

        glColor4f(1, 1, 1, elapsed/40);
        glBegin(GL_LINES);

        glVertex2f(x1-off, y1);
        glVertex2f(x2-off, y1);

        glVertex2f(x2, y1-off);
        glVertex2f(x2, y2-off);

        glVertex2f(x2+off, y2);
        glVertex2f(x1+off, y2);

        glVertex2f(x1, y2+off);
        glVertex2f(x1, y1+off);

        glEnd();

        drawStringC(msg, 240, 260, 14, 18, 1,1,1, elapsed/30);
    }

    private void doYesNo(char[] msg) {
        glDisable(GL_TEXTURE_2D);

        float elapsed = Sino.obj.timer.getTurn - stateStartTurn_;
        if (elapsed > 30) elapsed = 30;

        float x1 = 50;
        float x2 = 430;
        float y1 = 210;
        float y2 = 290;

        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glColor4f(0, 0, 0, elapsed/35);
        glBegin(GL_QUADS);
        glVertex2f(x1, y1);
        glVertex2f(x2, y1);
        glVertex2f(x2, y2);
        glVertex2f(x1, y2);
        glEnd();
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        float off = 30-elapsed;

        glColor4f(1, 1, 1, elapsed/40);
        glBegin(GL_LINES);

        glVertex2f(x1-off, y1);
        glVertex2f(x2-off, y1);

        glVertex2f(x2, y1-off);
        glVertex2f(x2, y2-off);

        glVertex2f(x2+off, y2);
        glVertex2f(x1+off, y2);

        glVertex2f(x1, y2+off);
        glVertex2f(x1, y1+off);

        glEnd();

        drawStringC(msg, 240, 245, 14, 18, 1,1,1, elapsed/30);
//                    240, 450, 14, 18, 1,1,1, elapsed/30);

        drawStringC("Yes.", 180, 280, 14, 18, 1,1,1, elapsed/40);
        drawStringC(" No.", 300, 280, 14, 18, 1,1,1, elapsed/40);
        if (yes_) drawStringC("Yes.", 181, 281, 14, 18, 1,0,0, elapsed/40);
        else drawStringC(" No.", 301, 281, 14, 18, 1,0,0, elapsed/40);

        if (Pad_GetAxis(0) & PAD_AXIS_RIGHT) {
            yes_ = false;
        }
        else if (Pad_GetAxis(0) & PAD_AXIS_LEFT) {
            yes_ = true;
        }
    }

    void doSaveReplay() {
        doYesNo("Save replay?");

        if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
            if (yes_) {
                state_ = State.NAME;
            }
            else {
                isEnd_ = true;
            }
        }
    }

    void doSendingReplay() {
        doMsg("sending...");

        if (httpThread.isEnd()) {
            HttpResponse res = httpThread.response;

            Log.info(`score server status: ` ~ res.statusLine);
            status_ = strip(res.messageBody);
            Log.info(status_);
        }
        else if (httpThread.isError()) {
            Log.info("err");
            status_ = `ERROR`;
        }
        else if (httpThread.isTimeout()) {
            Log.info("timeout");
            status_ = `TIMEOUT`;
        }
        else {
            return;
        }

        state_ = State.SENDED;
        stateStartTurn_ = Sino.obj.timer.getTurn();
    }

    void doSendedReplay() {
        doMsg(status_);

        if (stateStartTurn_+90 < Sino.obj.timer.getTurn()) {
            isEnd_ = true;
        }
    }

    void doSendReplay() {
        doYesNo("Send replay?");

        if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
            if (yes_) {
                char[] q = cast(char[])std.file.read(`replay/` ~ fname_);
                httpThread = new HttpThread(q);
                httpThread.start();

                stateStartTurn_ = Sino.obj.timer.getTurn;
                state_ = State.SENDING;
            }
            else {
                isEnd_ = true;
            }
        }
    }

    void doNameEntry() {
        try {
            drawStringC("Name Entry", 240, 100, 14, 18, 1,1,1);

            char[][] sys;
            sys ~= "END";
            sys ~= "CLEAR";
            sys ~= "<";
            sys ~= ">";
            foreach (int i, char[] s; sys) {
                drawStringC(s, 120+i*80, 160, 12, 18, 1,1,1,0.7);
            }
            for (int j = 0; j < 6; j++) {
                for (int i = 0; i < 16; i++) {
                    char[] b;
                    b ~= cast(char)(j*16+i+32);
                    drawStringC(b, 90+i*20, 200+j*30, 14, 18, 1,1,1,0.7);
                }
            }

            if (ny_ == 0) {
                drawStringC(sys[nx0_], 120+nx0_*80+1, 160+1, 12, 18,
                            1,0,0,0.7);
            }
            else {
                char[] b;
                b ~= cast(char)(ny_*16+nx_+16);
                drawStringC(b, 90+nx_*20, 170+ny_*30, 14, 18, 1,0,0,0.7);
                drawStringC("_", 90+nx_*20, 175+ny_*30, 14, 18, 1,0,0,0.7);
            }

            foreach (int i, char c; name_) {
                char[] b;
                b ~= c;
                drawStringC(b, 110+i*20, 390, 14, 18, 1,1,1,0.7);
                drawStringC(b, 110+i*20+1, 390+1, 14, 18, 0,0,1,0.7);
            }
            drawStringC("_", 110+nc_*20, 395, 14, 18, 1,1,1,0.7);
            drawStringC("_", 110+1+nc_*20, 395+1, 14, 18, 0,0,1,0.7);

            if (state_ == State.SEND) {
                doSendReplay();
                return;
            }
            else if (state_ == State.SENDING) {
                doSendingReplay();
                return;
            }
            else if (state_ == State.SENDED) {
                doSendedReplay();
                return;
            }

            if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                if (nc_ == 0) {
                    if (name_.length > 0) {
                        name_ = name_[1..length];
                    }
                }
                else {
                    name_ = name_[0..nc_-1] ~ name_[nc_..length];
                    nc_--;
                }
            }
            if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
                if (ny_ == 0) {
                    if (nx0_ == 0) {
                        if (name_.length > 0) {
                            state_ = State.SEND;
                            yes_ = true;
                            stateStartTurn_ = Sino.obj.timer.getTurn;
                            saveReplayFile();
                            Sino.rc.set("name", name_);
                            Sino.rc.dump();
                        }
                    }
                    else if (nx0_ == 1) {
                        name_ = "";
                        nc_ = 0;
                    }
                    else if (nx0_ == 2) {
                        if (nc_ > 0) nc_--;
                    }
                    else if (nx0_ == 3) {
                        if (nc_ < name_.length) nc_++;
                    }
                }
                else {
                    char c = cast(char)(ny_*16+nx_+16);
                    if (nc_ < name_.length) {
                        name_[nc_] = c;
                    }
                    else {
                        name_ ~= c;
                    }
                    if (nc_ < 13) nc_++;
                }
            }

            static int wait_;
            wait_--;
            if (wait_ < 0) {
                int axis = Pad_GetAxis(0);

                if (axis != PAD_AXIS_NONE) wait_ = 10;

                if ((axis & PAD_AXIS_UP) && ny_ > 0) {
                    ny_--;
                }
                else if ((axis & PAD_AXIS_DOWN) && ny_ < 6) {
                    ny_++;
                }
                if (ny_ == 0) {
                    if ((axis & PAD_AXIS_LEFT) && nx0_ > 0) {
                        nx0_--;
                    }
                    else if ((axis & PAD_AXIS_RIGHT) && nx0_ < 3) {
                        nx0_++;
                    }
                }
                else {
                    if ((axis & PAD_AXIS_LEFT) && nx_ > 0) {
                        nx_--;
                    }
                    else if ((axis & PAD_AXIS_RIGHT) && nx_ < 15) {
                        nx_++;
                    }
                }
            }
        }
        catch (Object o) {
            Log.warn("exception in name entry: " ~ o.toString());
        }
    }

    void saveReplayFile() {
        char[] formatDate(d_time d) {
            int sec = SecFromTime(d);
            int min = MinFromTime(d);
            int hour = HourFromTime(d);
            int date = DateFromTime(d);
            int month = MonthFromTime(d)+1;
            int year = YearFromTime(d);
            year %= 100;
            char[] tos(int i) {
                if (i < 10) return "0"~tostr(i);
                else return tostr(i);
            }
            return tos(year)~tos(month)~tos(date)~tos(hour)~tos(min)~tos(sec);
        }

        d_time d = getUTCtime();
        char[] dt = formatDate(d);
        char[] encname = fnEncode(name_);
        fname_ = dt ~ "_" ~ encname ~ ".rpy";
        Log.info("replay save to replay/" ~ fname_);

        File ofile = new File("replay/" ~ fname_, FileMode.OutNew);
        with (result_) {
            ofile.writeLine("filename: " ~ fname_);
            ofile.writeLine("version: " ~ REPLAY_VERSION);
            ofile.writeLine("score: " ~ tostr(allScore));
            ofile.writeLine("name: " ~ name_);
            ofile.writeLine("date: " ~ dt);
            ofile.writeLine("utcTime: " ~ tostr(d));
            ofile.writeLine("gameScore: " ~ tostr(gameScore));
            ofile.writeLine("turnScore: " ~ tostr(turnScore));
            ofile.writeLine("clearScore: " ~ tostr(clearScore));
            ofile.writeLine("leftScore: " ~ tostr(leftScore));
            ofile.writeLine("nomissScore: " ~ tostr(nomissScore));
            ofile.writeLine("scoreStatEnc: divPrintable95");
            int prev = 0;
            char[] scoreStatEnc = "";
            foreach (int s; scoreStat) {
                int d = s - prev;
                for (int i = 0; i < 4; i++) {
                    scoreStatEnc ~= cast(char)(d % 95 + 32);
                    d /= 95;
                }
                prev = s;
            }
            ofile.writeString("scoreStat: " ~ scoreStatEnc);
/* not use CSV
            foreach (int s; scoreStat) {
                ofile.writeString(tostr(s)~",");
            }
*/
        }
        ofile.writeLine("");

        game_.saveReplay(ofile);
        ofile.close();

        Log.info("replay save end");
    }

private:
    Game game_;

    bool isEnd_;

    enum State {
        RESULT, SAVE, NAME, SEND, SENDING, SENDED
    }
    State state_;
    int stateStartTurn_;
    bool yes_;
    char[] status_;

    char[] name_;
    int nx_, nx0_, ny_, nc_;

    ResultDisplay result_;

    char[] fname_;

}

