private import scene;
private import common;
private import sino;
private import rc;
private import mygl;
private import myglbf;
private import game;
private import replaydb;
private import result_display;
private import sound;

private import SDL_pad;

private import std.string;
private import std.date;

class SceneSelect : Scene {
public:
    this() {
        Log.info("scene select");
        isEnd_ = false;

        origNoSound_ = Sino.rc.get(`noSound`);

        isTitleWithSample_ = Sino.rc.getBit(`titleWithSample`);
        if (isTitleWithSample_) {
            Sino.rc.set(`noSound`, `1`);
            game_ = new Game(`data/demo.rpy`);
            Sino.rc.set(`noSound`, origNoSound_);
            startWait_ = 300;
        }

        h_ = 1;
        demoView_ = false;

        state_ = State.MENU;

        result_ = new ResultDisplay(`Replay`);

        replayMode_ = 1;
    }

    bit isEnd() {
        return isEnd_;
    }

    bool isEndIsQuit() {
        return state_ == State.MENU;
    }

    bool isDemoRunning() {
        return isTitleWithSample_ && startWait_ < 0;
    }

    Scene getNext() {
        return new SceneGame();
    }

    void run() {
        if (state_ == State.REPLAY_RUN) {
            game_.run();

            drawString(`press DASH button to exit`,
                       10, 470, 8, 10, 1,1,1,0.7);
            if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                game_.end();
            }

            if (game_.isEnd()) {
                Sound.obj.fadeMusic();
                state_ = State.REPLAY;
            }
            return;
        }

        glClear(GL_COLOR_BUFFER_BIT);

        if (state_ == State.REPLAY || state_ == State.REPLAY_MODE ||
            state_ == State.REPLAY_PREPARE)
        {
            draws_.draw();
            doReplay();
            return;
        }

        startWait_--;
        if (isDemoRunning) {
            game_.run();
            if (game_.isEnd()) {
                delete game_;
                Sino.rc.set(`noSound`, `1`);
                game_ = new Game(`data/demo.rpy`);
                Sino.rc.set(`noSound`, origNoSound_);
                startWait_ = 300;
            }

            if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                demoView_ = !demoView_;
            }
            if (demoView_) {
                drawString(`press DASH button to return menu`,
                           10, 470, 8, 10, 1,1,1,0.7);
                return;
            }
            else {
                glDisable(GL_TEXTURE_2D);
                glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                glColor4f(0, 0, 0, 0.5);
                glBegin(GL_QUADS);
                glVertex2f(480, 0);
                glVertex2f(480, 480);
                glVertex2f(0, 480);
                glVertex2f(0, 0);
                glEnd();
                glBlendFunc(GL_SRC_ALPHA, GL_ONE);

                drawString(`press DASH button to view demo play`,
                           10, 470, 8, 10, 1,1,1,0.7);
            }
        }
        else {
            draws_.draw();
        }

        if (state_ == State.MENU) {
            doMenu();
        }
    }

    void setResultDisplay() {
        if (replayDB_.sorted.length == 0) {
            return;
        }
        if (replayIte_ >= replayDB_.sorted.length) {
            replayIte_ = 0;
        }

        void copyEntry(ReplayEntry src, inout ResultDisplay dst) {
            dst.gameScore = src.gameScore;
            dst.turnScore = src.turnScore;
            dst.clearScore = src.clearScore;
            dst.leftScore = src.leftScore;
            dst.nomissScore = src.nomissScore;
            dst.allScore = src.allScore;

            dst.endTurn = src.endTurn;
            dst.scoreStat.length = src.scoreStat.length;
            dst.scoreStat[] = src.scoreStat;

            dst.hasReplayInfo = true;
            dst.mode = src.mode;
            dst.pversion = src.pversion;
            dst.name = src.name;
            dst.date = src.date;
        }
        copyEntry(replayDB_.sorted[replayIte_], result_);
/*
        copyEntry!(ReplayDB.Entry, ResultDisplay)
            (replayDB_.sorted[replayIte_], result_);
*/
        result_.init();
    }

    private void doMsg(char[] msg) {
        glDisable(GL_TEXTURE_2D);

        float elapsed = Sino.obj.timer.getTurn - stateStartTurn_;
        if (elapsed > 30) elapsed = 30;

        float x1 = 50;
        float x2 = 430;
        float y1 = 230;
        float y2 = 270;

        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glColor4f(0, 0, 0, elapsed/35);
        glBegin(GL_QUADS);
        glVertex2f(x1, y1);
        glVertex2f(x2, y1);
        glVertex2f(x2, y2);
        glVertex2f(x1, y2);
        glEnd();
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        float off = 30-elapsed;

        glColor4f(1, 1, 1, elapsed/40);
        glBegin(GL_LINES);

        glVertex2f(x1-off, y1);
        glVertex2f(x2-off, y1);

        glVertex2f(x2, y1-off);
        glVertex2f(x2, y2-off);

        glVertex2f(x2+off, y2);
        glVertex2f(x1+off, y2);

        glVertex2f(x1, y2+off);
        glVertex2f(x1, y1+off);

        glEnd();

        drawStringC(msg, 240, 260, 14, 18, 1,1,1, elapsed/30);
    }

    void doReplay() {
        drawString(`Select mode`, 50, 50, 10, 12, 1, 1, 1, 0.9);
        char[][] modes;
        modes ~= `easy`;
        modes ~= `normal`;
        modes ~= `endless`;
        foreach (int i, char[] m; modes) {
            drawString(m, 100, 80+25*i, 10, 12, 1, 1, 1, 0.9);
        }
        drawString(modes[replayMode_], 101,
                   81+25*replayMode_, 10, 12, 1, 0, 0, 0.9);

        if (state_ != State.REPLAY_PREPARE) {
            if (!replayDB_.isOk()) {
                doMsg(`loading replay...`);
                return;
            }

            if (replayDB_.isError()) {
                Log.warn(`no replay in replay directory`);
                state_ = State.MENU;
                return;
            }
        }

        if (state_ == State.REPLAY_MODE) {
            static int wait_;
            wait_--;
            if (wait_ < 0) {
                int axis = Pad_GetAxis(0);

                if (axis != PAD_AXIS_NONE) wait_ = 10;

                if ((axis & PAD_AXIS_UP) && replayMode_ > 0) {
                    replayMode_--;
                }
                else if ((axis & PAD_AXIS_DOWN) &&
                         replayMode_ < modes.length-1)
                {
                    replayMode_++;
                }
            }

            if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
                replayDB_.setSort(modes[replayMode_], this);
                replayIte_ = 0;
                state_ = State.REPLAY;
                stateStartTurn_ = Sino.obj.timer.getTurn;
            }
            else if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                state_ = State.MENU;
            }

            return;
        }

        if (replayDB_.sorted.length == 0) {
            state_ = State.REPLAY_MODE;
            return;
        }

        for (int i = 0; i < 10; i++) result_.move();

        glPushMatrix();
        glTranslatef(240, 15, 0);
        glScalef(0.5, 0.5, 1);
        result_.draw();
        glPopMatrix();

        int replayNum = replayDB_.sorted.length;

        void drawReplayEntry(ReplayEntry e, float y,
                             float r, float g, float b) {
            d_time d = UTCtoLocalTime(e.utcTime);
            int date = DateFromTime(d);
            int month = MonthFromTime(d)+1;
            int year = YearFromTime(d);
            char[] dt = format("%2d/%2d/%2d", year, month, date);
            drawString(format("%-14s   %s   %s   %8d",
                              e.name, e.pversion, dt, e.allScore),
                       50, y, 10, 12, r, g, b, 0.9);
        }

        drawString(`Name             Ver   Date         Score`,
                   45, 265, 10, 12, 1, 1, 1, 0.9);
        int replayPage = replayIte_ / REPLAY_VIEW_NUM;
        int replayOffset = replayPage * REPLAY_VIEW_NUM;
        for (int i = 0; i < REPLAY_VIEW_NUM; i++) {
            int rn = i + replayOffset;
            if (rn >= replayNum) break;
            drawReplayEntry(replayDB_.sorted[rn], 240+50+i*15, 1,1,1);
        }
        drawReplayEntry(replayDB_.sorted[replayIte_],
                        240+51+(replayIte_-replayOffset)*15, 1,0,0);

        if (state_ == State.REPLAY_PREPARE) {
            if (!replayDB_.isOk()) {
                doMsg(`getting replay file...`);
                return;
            }

            if (replayDB_.isError()) {
                Log.warn(`cannot get replay`);
                state_ = State.REPLAY;
                return;
            }

            game_ = new Game(rpyfile_);
            state_ = State.REPLAY_RUN;
        }

        if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
            state_ = State.REPLAY_PREPARE;
            stateStartTurn_ = Sino.obj.timer.getTurn;
            rpyfile_ = `replay/`~replayDB_.sorted[replayIte_].filename;
            replayDB_.prepareReplayFile(replayDB_.sorted[replayIte_]);
            return;
        }
        else if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
            state_ = State.REPLAY_MODE;
        }

        static int wait2_;
        wait2_--;
        if (wait2_ < 0) {
            int axis = Pad_GetAxis(0);

            if ((axis & PAD_AXIS_UP) && replayIte_ > 0) {
                replayIte_--;
            }
            else if ((axis & PAD_AXIS_DOWN) && replayIte_ < replayNum-1) {
                replayIte_++;
            }
            if (axis & PAD_AXIS_LEFT) {
                replayIte_ -= REPLAY_VIEW_NUM;
                if (replayIte_ < 0) replayIte_ = 0;
            }
            else if (axis & PAD_AXIS_RIGHT) {
                replayIte_ += REPLAY_VIEW_NUM;
                if (replayIte_ >= replayNum) replayIte_ = replayNum-1;
            }

            if (axis != PAD_AXIS_NONE) {
                wait2_ = 10;
                setResultDisplay();
            }
        }
    }

    void doMenu() {
        char[][] menu;
        menu ~= `EASY MODE`;
        menu ~= `NORMAL MODE`;
        menu ~= `ENDLESS MODE`;
        menu ~= `REPLAY`;
        menu ~= `NET REPLAY`;

        static int wait_;
        wait_--;
        if (wait_ < 0) {
            int axis = Pad_GetAxis(0);

            if (axis != PAD_AXIS_NONE) wait_ = 10;

            if ((axis & PAD_AXIS_UP) && h_ > 0) {
                h_--;
            }
            else if ((axis & PAD_AXIS_DOWN) && h_ < menu.length-1) {
                h_++;
            }
        }

        foreach (int i, char[] m; menu) {
            drawStringC(m, 240, 300+i*25, 12, 12, 1,1,1,0.7);
        }
        drawStringC(menu[h_], 241, 301+h_*25, 12, 12, 1,0,0,0.7);

        if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
            if (h_ < 3) {
                isEnd_ = true;
                Sino.rc.set(`noSound`, origNoSound_);

                if (h_ == 0) {
                    Sino.rc.set(`gameMode`, `easy`);
                }
                else if (h_ == 1) {
                    Sino.rc.set(`gameMode`, `normal`);
                }
                else if (h_ == 2) {
                    Sino.rc.set(`gameMode`, `endless`);
                }
            }
            else if (h_ == 3) {
                replayDB_ = new LocalReplayDB();
                state_ = State.REPLAY_MODE;
                stateStartTurn_ = Sino.obj.timer.getTurn;
            }
            else if (h_ == 4) {
                replayDB_ = new NetReplayDB();
                state_ = State.REPLAY_MODE;
                stateStartTurn_ = Sino.obj.timer.getTurn;
            }
        }

/*
        drawString("press A button to play normal game.",
                   10, 200, 14, 18, 1,1,1);
        drawString("press B button to view sample play.",
                   10, 230, 14, 18, 1,1,1);

        if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
            isEnd_ = true;
            Sino.rc.set(`noSound`, origNoSound_);
        }
        else if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
            isEnd_ = true;
            Sino.rc.set(`noSound`, origNoSound_);
        }
*/
    }

    int objectNum() {
        if (isDemoRunning) {
            return game_.objectNum();
        }
        else {
            return 0;
        }
    }

private:
    bool isEnd_;
    Game game_;
    bool isTitleWithSample_;
    int startWait_;

    char[] origNoSound_;

    int h_;

    bool demoView_;

    enum State {
        MENU, REPLAY_MODE, REPLAY, REPLAY_PREPARE, REPLAY_RUN
    }
    State state_;
    int stateStartTurn_;

    ReplayDB replayDB_;
    int replayIte_;

    ResultDisplay result_;
    int replayMode_;

    char[] rpyfile_;

    static const int REPLAY_VIEW_NUM = 12;

};

