private import std.string;
private import std.format;
private import std.date;
private import std.random;
private import std.gc;
private import std.conv;

private import SDL;
private import SDL_net;
private import SDL_pad;
private import opengl;
private import openglu;

private import rc;
private import log;
private import fpstimer;
private import scene;
private import scene_select;
private import myio;
private import myglbf;
private import sound;
private import texture;
private import http;

class Sino {
private:
    static Sino obj_;
public:
    static Sino obj() {
        return obj_;
    }

public:
    this(char args[][]) {
        obj_ = this;
        Log.info("init");
        initResource(args);
        initMisc();
        initSDL();
        initGL();
        Log.info("init done");
    }

    ~this() {
        Log.info("quit");
        delete sound_;
        SDLNet_Quit();
        SDL_Quit();
        Log.info("quit done");
    }

    void initResource(char args[][]) {
        Log.info("init resource");

        rc_ = new Rc();
        rc().addResourceFile();
        rc().addArgument(args);
    }

    void initMisc() {
        Log.info("init misc");

        rand_seed(getUTCtime(), 0);

        timer_ = new FpsTimer(rc.getBit("timerEnable"));
        int fps = rc.getInt("timerFps");
        timer_.setFps(fps);

        sound_ = new Sound();

        end_ = false;

        Proxy.host = rc.get(`proxyServer`);
        if (rc.get(`proxyPort`) != ``) {
            Proxy.port = rc.getInt(`proxyPort`);
        }
        Proxy.user = rc.get(`proxyUser`);
        Proxy.pass = rc.get(`proxyPass`);
    }

    void initSDL() {
        Log.info("init SDL");

        int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);
        if (ret != 0) Log.errorSDL("cannot init SDL");

        ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
        if (ret != 0) {
            Log.warnSDL("cannot init joystick");
        }

        useAudio_ = true;
        ret = SDL_InitSubSystem(SDL_INIT_AUDIO);
        if (ret != 0) {
            Log.warnSDL("cannot init audio");
            useAudio_ = false;
        }

        SDLNet_Init();

        SDL_VideoInfo* info = SDL_GetVideoInfo();

        int bpp = Sino.rc().getInt("screenBpp");
        char[] bppMsg;

        if (bpp == -1) {
            bpp = info.vfmt.BitsPerPixel;
            Log.info("using the \"best\" bpp: " ~ tostr(bpp));
        }
        else if (bpp == 0) {
            bpp = info.vfmt.BitsPerPixel;
            Log.info("using current bpp");
        }
        else {
            Log.info("using user set bpp: " ~ tostr(bpp));
        }

        SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );  
        SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
        SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );  
//        SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
        SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

        char[][] screenSize = Sino.rc.get(`screenSize`).split(`x`);
        screenWidth_ = toInt(screenSize[0]);
        screenHeight_ = toInt(screenSize[1]);
        int screenFull = Sino.rc().getInt("screenFull");
        int flags = SDL_OPENGL;
        if (screenFull) {
            flags |= SDL_FULLSCREEN;
        }

        Log.info("set video mode " ~
                 "size = " ~ tostr(screenWidth_) ~
                 "x" ~ tostr(screenHeight_) ~
                 ", bpp = " ~ tostr(bpp) ~
                 ", flags = " ~ tostr(flags));
        if (SDL_SetVideoMode(screenWidth_, screenHeight_,
                             bpp, flags) == null)
        {
            Log.errorSDL("cannot create screen");
        }

        info = SDL_GetVideoInfo();
        Log.info("set bpp: " ~ tostr(cast(int)info.vfmt.BitsPerPixel));

        Pad_Init();
        Pad_LoadDefault("save/key.conf");

        if (flags & SDL_FULLSCREEN) {
            SDL_ShowCursor(SDL_DISABLE);
        }

        SDL_WM_SetCaption(`Psy-NO`, null);
    }

    void initGL() {
        glClearColor(0, 0, 0, 0);
        glViewport(0, 0, screenWidth_, screenHeight_);

        glEnable(GL_BLEND);
        if (rc.getBit(`lineAntiAlias`)) {
            glEnable(GL_LINE_SMOOTH);
        }

        glBlendFunc(GL_SRC_ALPHA, GL_ONE);

        glMatrixMode(GL_MODELVIEW);

        // from PARSEC47
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_COLOR_MATERIAL);

        Texture.initTextures();

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        if (glGetError()) {
            Log.error(tostr(gluErrorString(glGetError())));
        }

        initLetter();
    }

    void run() {
        Log.info("run");

        Scene.init();
        scene_ = new SceneSelect();

        timer_.start();

        while (scene_ !== null) {
            Pad_Pump(PAD_PUMP_EVERYTHING);

            scene_.run();
            if (!timer_.shouldSkip()) {
                drawSystemInfo();
                SDL_GL_SwapBuffers();
            }
            if (scene_.isEnd()) {
                Scene prevScene = scene_;
                scene_ = scene_.getNext();
                delete prevScene;
                timer_.start();
            }
/*
            else if (!(timer_.getTurn() % 62)) {
                char[] msg =
                    "FPS: " ~ tostr(timer_.getSecFps()) ~
                    "  TURN: " ~ tostr(timer_.getTurn());
                if (scene_.objectNum() != -1) {
                    msg ~=
                        "  OBJS: " ~ tostr(scene_.objectNum());
                }
                msg ~= "  DROPS: " ~ tostr(timer_.getDroppedTurn());
                Log.info(msg);
            }
*/
            timer_.wait();
            endCheck();
        }

        Log.info("run done");
    }

    void drawSystemInfo() {
        char[] desc = "FPS / OBJ / DROP / TURNS";
        drawString(desc, 490, 465, 7, 12, 1,1,1);
        char[] info = format("%3d / %3d / %4d / %5d",
                             timer_.getSecFps(),
                             scene_.objectNum(),
                             timer_.getDroppedTurn(),
                             timer_.getTurn());
        drawString(info, 490, 475, 7, 12, 1,1,1);
/*
        int offx1 = 535;
        int offx2 = 620;
        int offy = 255;
        drawString("FPS", offx1, offy, 12, 16, 1,1,1);
        drawNumR(timer_.getSecFps(), offx2, offy+25, 12, 16, 1,1,1);
        drawString("OBJS", offx1, offy+55, 12, 16, 1,1,1);
        drawNumR(scene_.objectNum(), offx2, offy+80, 12, 16, 1,1,1);
        drawString("TURNS", offx1, offy+110, 12, 16, 1,1,1);
        drawNumR(timer_.getTurn(), offx2, offy+135, 12, 16, 1,1,1);
        drawString("FRM DROPS", offx1, offy+165, 10, 16, 1,1,1);
        drawNumR(timer_.getDroppedTurn(), offx2, offy+190, 12, 16, 1,1,1);
*/
/*
        drawString("SYS DROPS", offx1, offy+220, 10, 16, 1,1,1);
        drawNumR(timer_.getSystemDroppedTurn(), offx2, offy+245, 12, 16, 1,1,1);
*/
    }

    void endCheck() {
        if (Pad_GetButtonPressed(PAD_END_BUTTON)) {
            Log.info("quit event catched");
            quit();
        }
        if (Pad_GetButtonPushed(PAD_C_BUTTON)) {
            if (scene_.isEndIsQuit()) {
                quit();
            }
            else {
                Sound.obj.fadeMusic();
                scene_ = new SceneSelect();
            }
        }
        if (end_) quit();
    }

    void quit() {
        Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
        scene_ = null;
    }

    void end() {
        Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
        scene_ = scene_.getNext();
    }

    void setEnd() {
        // @@@ temporary...
        end_ = true;
    }

public:
    static Rc rc() { return obj().rc_; }

    bit useAudio() { return useAudio_; }

    FpsTimer timer() { return timer_; }

    int screenWidth() { return screenWidth_; }
    int screenHeight() { return screenHeight_; }

private:
    Rc rc_;

    FpsTimer timer_;

    bit useAudio_;

    Scene scene_;

    int screenWidth_, screenHeight_;

    Sound sound_;

    bool end_;

}


extern(C) int doit(char args[][]) {
    Log.init();
    Sino c = new Sino(args);
    c.run();
    delete c;
    Log.quit();

    return 0;
}

version (Windows) {
    // boot from winmain.d
}
else {
    int main(char args[][]) {
        return doit(args);
    }
}

