private import std.string;

private import SDL_mixer;

private import sino;
private import rc;
private import log;
private import sound_phase;

enum Se {
    PLAYER_DEAD,
    PLAYER_EXTEND,
    SMALL_ENEMY_DESTROY,
    ENEMY_APPEAR,
    SE_NUM
}

/// from PARSEC47
class Sound {
private:
    static const char[][Se.SE_NUM] seNames_ = [
        Se.PLAYER_DEAD:"sePlayerDead",
        Se.PLAYER_EXTEND:"sePlayerExtend",
        Se.SMALL_ENEMY_DESTROY:"seSmallEnemyDestroy",
        Se.ENEMY_APPEAR:"seEnemyAppear",
        ];

    static const int[Se.SE_NUM] seChannels_ = [
        Se.PLAYER_DEAD:0,
        Se.PLAYER_EXTEND:1,
        Se.SMALL_ENEMY_DESTROY:3,
        Se.ENEMY_APPEAR:4,
        ];

private:
    bool noSound_ = false;
    int fadeOutSpeed_ = 1280;

    static Sound obj_;

public:
    static Sound obj() { return obj_; }

public:
    static const char[] soundsDir = "data/sound/";
    static const char[] chunksDir = "data/sound/";

    void init() {
        noSound_ = Sino.rc.getBit("noSound");
    }

    this() {
        obj_ = this;

        init();

        if (noSound_) return;

        int audio_rate;
        Uint16 audio_format;
        int audio_channels;
        int audio_buffers;

        if (SDL_InitSubSystem(SDL_INIT_AUDIO) < 0) {
            noSound_ = 1;
            Log.warnSDL("Unable to initialize SDL_AUDIO");
            return;
        }

        audio_rate = 44100;
        audio_format = AUDIO_S16;
        audio_channels = 1;
        audio_buffers = 4096;
        if (Mix_OpenAudio(audio_rate, audio_format,
                          audio_channels, audio_buffers) < 0) {
            noSound_ = 1;
            Log.warnSDL("Couldn't open audio");
            return;
        }
        Mix_QuerySpec(&audio_rate, &audio_format, &audio_channels);

        loadChunks();

        int soundVolume = Sino.rc.getInt("soundVolume")*128/100;
        int seVolume = Sino.rc.getInt("seVolume")*128/100;
        Mix_VolumeMusic(soundVolume);
        for (int i = 0; i < Se.SE_NUM; i++) {
            if (chunk_[i]) Mix_VolumeChunk(chunk_[i], seVolume);
        }
    }

    void close() {
        if (noSound_) return;
        if (Mix_PlayingMusic()) {
            Mix_HaltMusic();
        }
        Mix_CloseAudio();
    }

    // Free a music or a chunk.
    ~this() {
        if (music_) {
            stopMusic();
            Mix_FreeMusic(music_);
            music_ = null;
        }
        foreach (Mix_Chunk* c; chunk_) {
            if (c) {
                Mix_FreeChunk(c);
                c = null;
            }
        }
        for (int i = 0; i < MIX_CHANNELS; i++) {
            haltChunk(i);
        }
    }

    // Load a sound or a chunk.

    void loadMusic(char[] name) {
        if (noSound_) return;

        if (music_) {
            stopMusic();
            Mix_FreeMusic(music_);
            music_ = null;
        }

        char[] fileName = soundsDir ~ name;
        music_ = Mix_LoadMUS(toStringz(fileName));
        if (!music_) {
//            noSound_ = true;
            Log.warnSDL("Couldn't load: " ~ fileName);
        }
    }
  
    // Play/Stop the music/chunk.

    void playMusic(int loop) {
        if (!noSound_) {
            if (music_) {
                Mix_PlayMusic(music_, loop);
            }
        }

        SoundPhase.init();
    }

    void fadeMusic() {
        if (noSound_) return;
        Mix_FadeOutMusic(fadeOutSpeed_);
    }

    void stopMusic() {
        if (noSound_) return;
        if (Mix_PlayingMusic() ) {
            Mix_HaltMusic();
        }
    }

    void playChunk(Se id) {
        if (noSound_) return;
        if (chunk_[id]) {
            Mix_PlayChannel(seChannels_[id], chunk_[id], 0);
        }
    }

    void haltChunk(int ch) {
        if (noSound_) return;
        Mix_HaltChannel(ch);
    }
    void haltChunk(Se id) {
        if (noSound_) return;
        Mix_HaltChannel(seChannels_[id]);
    }

private:
    void loadChunks() {
        if (noSound_) return;

        for (int i = 0; i < Se.SE_NUM; i++) {
            char[] seName = seNames_[i];
            char[] seFileName = chunksDir ~ Sino.rc.get(seName);
            chunk_[i] = Mix_LoadWAV(toStringz(seFileName));
            if (!chunk_[i]) {
                Log.warnSDL("Couldn't load: " ~ seFileName);
            }
        }
    }

private:
    Mix_Music* music_;
    Mix_Chunk*[Se.SE_NUM] chunk_;

}

/**
@memo
usage image ...

Sound.obj.playMusic(1);
Sound.obj.fadeMusic();
Sound.obj.haltMusic();
Sound.obj.playChunk(PLAYER_SHOT_A);
Sound.obj.haltChunk(PLAYER_SHOT_A);

*/

